"""
Utilities for setting up Python's weirdo logging system. Including logging to
MongoDB if you feel crazy.
"""

from __future__ import annotations

import atexit
import logging
import logging.handlers
import queue
import socket
import threading
import time
import weakref
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import *  # type: ignore

import uniserde
from bson import ObjectId
from uniserde import BsonDoc

from . import async_utils

try:
    import pymongo.collection
except ImportError:
    if TYPE_CHECKING:
        import pymongo.collection


__all__ = [
    "LogLevel",
    "LOG_LEVEL_NAMES",
    "LOG_LEVEL_ANNOUNCEMENT",
    "Environment",
    "levels_greater_than_or_equal",
    "MongoDbLogger",
    "setup_logging",
]


T = TypeVar("T")


LogLevel: TypeAlias = Literal[
    "debug",
    "info",
    "announcement",
    "warning",
    "error",
    "fatal",
]
LOG_LEVEL_NAMES = get_args(LogLevel)  # In ascending order!

# Add the announcement level to the logging module
LOG_LEVEL_ANNOUNCEMENT = (logging.INFO + logging.WARNING) // 2
logging.addLevelName(LOG_LEVEL_ANNOUNCEMENT, "ANNOUNCEMENT")

Environment: TypeAlias = Literal["development", "production"]


# Keep track of all living persistence loggers, so they can be flushed at exit
_ALL_MONGODB_LOGGERS: weakref.WeakSet[MongoDbLogger] = weakref.WeakSet()


@atexit.register
def _flush_all_persistence_loggers() -> None:
    for logger in _ALL_MONGODB_LOGGERS:
        logger.flush_sync()


def _log_level_to_python(level: LogLevel) -> int:
    return logging.getLevelName(level.upper())


def _log_level_from_python(level: int) -> LogLevel:
    return logging.getLevelName(level).lower()


def levels_greater_than_or_equal(level: LogLevel) -> set[LogLevel]:
    """
    Returns a set containing all log levels which are at least as severe as the
    given one.

    ## Raises

    `ValueError`: if the given level is not a valid log level.
    """
    for ii, name in enumerate(LOG_LEVEL_NAMES):
        if name == level:
            return set(LOG_LEVEL_NAMES[ii:])

    raise ValueError(f"Unknown log level: {level}")


@dataclass
class LogEntry(uniserde.Serde):
    """
    Represents a single log entry in the database.
    """

    # Unique identifier for the log entry
    id: ObjectId

    # When this log entry was created
    timestamp: datetime

    # Whether this log entry was created in development or production
    environment: Environment

    # Which machine created this log entry
    host: str

    # The app that created this log entry
    app: str

    # The severity of this log entry
    level: LogLevel

    # A human-readable message. This is the main content of the log entry
    message: str

    # Arbitrary additional data
    payload: dict[str, Any]


class MongoDbLogger(logging.StreamHandler):
    """
    Logger, which stores its entries in a MongoDB database.

    Database access is asynchronous, and likely on another server. `await`ing
    every log operation would be incredibly slow. Instead, this logger only
    synchronously queues log entries, and then later asynchronously copies them
    to the database.
    """

    def __init__(
        self,
        collection: pymongo.collection.Collection,
        *,
        environment: Environment,
        app: str,
        host: str | None = None,
    ):
        super().__init__()

        self._sync_collection = collection

        self._host = socket.gethostname() if host is None else host
        self._environment: Environment = environment
        self._app = app

        self._last_log_writeback_time_monotonic = time.monotonic()
        self._pending_log_entries: queue.Queue[LogEntry] = queue.Queue()

        self._writeback_thread = threading.Thread(
            target=self._writeback_worker, daemon=True
        )
        self._writeback_thread.start()

        # Log entries are plentiful. Make sure there's an index
        collection.create_index([("timestamp", 1)])

        # Keep track of all living persistence loggers
        _ALL_MONGODB_LOGGERS.add(self)

    def __del__(self) -> None:
        self.flush_sync()

    def create_log_entries_sync(self, entries: Iterable[LogEntry]) -> None:
        """
        Batch creates new log entries in the database.

        The entries need to have unique ids among all entries in the database.
        This is not checked for performance reasons.
        """
        # MongoDB doesn't like empty inserts
        entry_data = [entry.as_bson() for entry in entries]

        if not entry_data:
            return

        # Insert the entries
        self._sync_collection.insert_many(entry_data)

    def flush_sync(self) -> None:
        """
        Copies any not yet stored log entries into the database.
        """
        # Move the log entries into a local variable
        in_flight_entries = []

        while True:
            try:
                in_flight_entries.append(self._pending_log_entries.get_nowait())
            except queue.Empty:
                break

        # Try to push the entries into the database
        try:
            self.create_log_entries_sync(in_flight_entries)

        # If the operation fails, put the entries back into the queue so they
        # can be retried later
        except Exception:
            for entry in in_flight_entries:
                self._pending_log_entries.put(entry)

            raise

    def _writeback_worker(self) -> None:
        # Keep copying entries forever. There is no need to return because the
        # thread is daemonized.
        #
        # TODO: What if the logger is deleted? This isn't currently supported.
        CYCLE_TIME = 10

        while True:
            # Wait some time before writing back. This batches log entries
            # together.
            now = time.monotonic()
            sleep_time = CYCLE_TIME - (now - self._last_log_writeback_time_monotonic)

            if sleep_time > 0:
                time.sleep(sleep_time)

            # Copy the pending log entries into the database
            try:
                self.flush_sync()
            except Exception as e:
                logging.error(f"Error writing back log entries: {e}")

                # Wait a bit before retrying
                time.sleep(20)

            # Housekeeping
            self._last_log_writeback_time_monotonic = time.monotonic()

    def queue_log(
        self,
        level: LogLevel,
        message: str,
        *,
        payload: dict[str, Any] | None = None,
    ) -> None:
        """
        Creates a log entry and queues it for storage to the database. The entry
        will be created later, or once `flush` is called.
        """
        self._pending_log_entries.put(
            LogEntry(
                id=ObjectId(),
                timestamp=datetime.now(timezone.utc),
                environment=self._environment,
                host=self._host,
                app=self._app,
                level=level,
                message=message,
                payload={} if payload is None else payload,
            )
        )

    def emit(self, record: logging.LogRecord) -> None:
        """
        For compatibility with python's logging module.
        """

        self.queue_log(
            level=_log_level_from_python(record.levelno),
            message=record.message,
        )

    def find_log_entries(
        self,
        *,
        environment: Environment | None = None,
        levels: Iterable[LogLevel] | None = None,
        newer_than: datetime | None = None,
        older_than: datetime | None = None,
        limit: int | None = None,
    ) -> AsyncIterable[LogEntry]:
        """
        Returns an async iterator over all log entries in the database matching
        the given filters.

        The entries are sorted by timestamp, with the most recent entries first.
        """
        # Build the query
        query: BsonDoc = {}

        if environment is not None:
            query["environment"] = uniserde.as_bson(environment)

        if levels is not None:
            query["level"] = {"$in": [uniserde.as_bson(level) for level in levels]}

        if newer_than is not None:
            query["timestamp"] = {"$gt": newer_than}

        if older_than is not None:
            query["timestamp"] = {"$lt": older_than}

        # Yield all matches
        def sync_iterator() -> Iterator[LogEntry]:
            cursor = self._sync_collection.find(query).sort([("timestamp", -1)])

            if limit is not None:
                cursor = cursor.limit(limit)

            for doc in cursor:
                yield LogEntry.from_bson(doc)

        return async_utils.iterator_to_thread(sync_iterator(), batch_size=50)

    def watch_log_entries(
        self,
        *,
        environment: Environment | None = None,
        levels: Iterable[LogLevel] | None = None,
    ) -> AsyncIterable[LogEntry]:
        """
        Returns an async iterator over all new log entries in the database
        matching the given filters.

        The iterator will block until new log entries are available.
        """
        # Build the filter pipeline
        query: BsonDoc = {"operationType": "insert"}

        if environment is not None:
            query["fullDocument.environment"] = uniserde.as_bson(environment)

        if levels is not None:
            query["fullDocument.level"] = {
                "$in": [uniserde.as_bson(level) for level in levels]
            }

        pipeline = [
            {
                "$match": query,
            }
        ]

        # Watch for changes
        def sync_iterator() -> Iterator[LogEntry]:
            for change in self._sync_collection.watch(pipeline=pipeline):
                yield LogEntry.from_bson(change["fullDocument"])

        return async_utils.iterator_to_thread(sync_iterator(), batch_size=1)


def reset_logging() -> None:
    """
    Resets the logging system to its initial state. This removes all handlers
    from the root logger and resets its level to `WARNING`.
    """

    # Get the root logger
    root_logger = logging.getLogger()

    # Remove all existing handlers
    for handler in root_logger.handlers[:]:
        root_logger.removeHandler(handler)

    # Reset the root logger level
    root_logger.setLevel(logging.WARNING)


@overload
def setup_logging(
    *,
    info_log_path: Path | None = None,
    debug_log_path: Path | None = None,
    stdout_log_level: LogLevel = "debug",
) -> MongoDbLogger:
    ...


@overload
def setup_logging(
    *,
    info_log_path: Path | None = None,
    debug_log_path: Path | None = None,
    database_collection: pymongo.collection.Collection | None = None,
    database_environment: Environment,
    database_connection_string: str | None = None,
    database_app: str,
    stdout_log_level: LogLevel = "debug",
    database_log_level: LogLevel = "debug",
) -> MongoDbLogger:
    ...


def setup_logging(
    *,
    info_log_path: Path | None = None,
    debug_log_path: Path | None = None,
    database_collection: pymongo.collection.Collection | None = None,
    database_environment: Environment | None = None,
    database_connection_string: str | None = None,
    database_app: str | None = None,
    stdout_log_level: LogLevel = "debug",
    database_log_level: LogLevel = "debug",
) -> MongoDbLogger | None:
    """
    Creates a nice logging setup. Any previously registered handlers are
    removed, which means it is safe to call this function multiple times without
    creating duplicate log entries.

    - INFO logs to `info_log_path`, keeping logs indefinitely
    - Configurable logs to `stdout`, DEBUG by default
    - DEBUG logs to `debug_log_path`, keeping a limited number of days
    - DEBUG logs into the database, if a collection is provided

    Returns the database logger, if one was created.

    ## Parameters

    `info_log_path`: Path to the info log file

    `debug_log_path`: Path to the debug log file

    `database_collection`: The MongoDB collection to store logs in

    `database_environment`: The environment to mark log entries as

    `database_connection_string`: The MongoDB connection string to connect to
        the database

    `database_app`: The running app's name. This will be added to logged
        entries in the database

    `stdout_log_level`: The minimum log level to display on stdout

    `database_log_level`: The minimum log level to store in the database
    """

    # Make sure not to add multiple handlers if the function is called multiple
    # times.
    reset_logging()

    # Configure the root logger and prepare some values
    root_logger = logging.getLogger("")
    root_logger.setLevel(logging.DEBUG)

    formatter = logging.Formatter("%(asctime)s  %(levelname)-8s  %(message)s")

    # Info -> file
    if info_log_path is not None:
        info_log_path.parent.mkdir(parents=True, exist_ok=True)

        handler = logging.handlers.TimedRotatingFileHandler(
            info_log_path,
            encoding="utf-8",
            when="midnight",
            utc=True,
        )
        handler.setLevel(logging.INFO)
        handler.setFormatter(formatter)
        root_logger.addHandler(handler)

    # Stdout (configurable)
    handler = logging.StreamHandler()
    handler.setLevel(_log_level_to_python(stdout_log_level))
    handler.setFormatter(formatter)
    root_logger.addHandler(handler)

    # Debug -> file
    if debug_log_path is not None:
        debug_log_path.parent.mkdir(parents=True, exist_ok=True)

        handler = logging.handlers.TimedRotatingFileHandler(
            debug_log_path,
            encoding="utf-8",
            when="midnight",
            utc=True,
            backupCount=7,
        )
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(formatter)
        root_logger.addHandler(handler)

    # Database
    if database_collection is None:
        pers_logger = None
    else:
        assert (
            database_environment is not None
        ), "Must provide an environment when logging to a database"

        assert (
            database_app is not None
        ), "Must provide an app name when logging to a database"

        pers_logger = MongoDbLogger(
            database_collection,
            host=database_connection_string,
            environment=database_environment,
            app=database_app,
        )

        pers_logger.setLevel(_log_level_to_python(database_log_level))
        root_logger.addHandler(pers_logger)

    return pers_logger



"""
Utilities for loading configuration from JSON files and environment variables.
"""

from __future__ import annotations

import enum
import os
from collections.abc import Iterable
from datetime import datetime, timedelta
from pathlib import Path
from typing import *  # type: ignore

import introspection.typing
import json5
import uniserde
import uniserde.case_convert
from typing_extensions import dataclass_transform
from uniserde import Jsonable, ObjectId

__all__ = [
    "ConfigurationError",
    "Config",
]


_DEFAULTS_BY_TYPE: dict[Type, uniserde.Jsonable] = {
    Any: None,
    bool: False,
    bytes: b"",
    datetime: datetime.now().isoformat(),
    dict: {},
    enum.Enum: "",
    enum.Flag: [],
    float: 0.0,
    int: 0,
    list: [],
    Literal: "",
    set: set(),
    str: "",
    timedelta: 0,
    tuple: tuple(),
    Union: None,
    ObjectId: "",
}  # type: ignore


class ConfigurationError(ValueError):
    """
    Raised when the configuration couldn't be loaded, for whichever reason.
    """

    pass


@dataclass_transform()
class Config:
    @classmethod
    def _get_fields(cls) -> Iterable[tuple[str, type]]:
        """
        Returns iterators over all fields and their types.
        """
        for attr_name, type_info in introspection.typing.get_type_annotations(
            cls
        ).items():
            if type_info.arguments:
                raw_type = introspection.typing.parameterize(
                    type_info.type, type_info.arguments
                )
            else:
                raw_type = type_info.type

            yield attr_name, raw_type  # type: ignore

    @staticmethod
    def _create_json_template(fields: Iterable[tuple[str, Type]]) -> str:
        """
        Create an empty JSON template for the provided fields.
        """

        # Build a JSON containing the default values for each field
        default_values_dict: dict[str, Jsonable] = {}

        for field_py_name, field_type in fields:
            field_doc_name = uniserde.case_convert.all_lower_to_camel_case(
                field_py_name
            )
            default_values_dict[field_doc_name] = _DEFAULTS_BY_TYPE.get(
                field_type, None
            )

        # Serialize the instance into formatted JSON
        serialized: str = json5.dumps(  # type: ignore
            default_values_dict,
            indent=4,
            quote_keys=True,
            trailing_commas=True,
        )

        # Add comments describing the fields
        serialized = serialized.strip()
        lines = serialized.splitlines()

        for ii, line in enumerate(lines[1:-1], start=1):
            lines[ii] = "  // " + line.strip()

        # Done
        return "\n".join(lines) + "\n"

    @classmethod
    def _parse_fields(
        cls,
        *,
        fields: Iterable[tuple[str, Type]],
        values: Iterable[tuple[str, Any]],
        case_transform: Callable[[str], str],
        raise_on_superfluous: bool,
    ) -> tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Parses the provided fields and values into a dictionary, transforming
        the field names using the provided `case_transform` function. The result
        is a tuple of two dictionaries: the first contains the parsed fields,
        the second any unparsed, i.e. superfluous values.

        ## Raises

        `ConfigurationError`: if any fields are missing, or of the wrong type.
        """
        # Prepare everything as dictionaries for quick lookup
        field_dict: dict[str, Type] = dict(fields)
        value_dict: dict[str, Any] = dict(values)

        # Parse the fields, popping them from the value dictionary
        parsed_fields: dict[str, Any] = {}

        for py_name, py_type in field_dict.items():
            doc_name = case_transform(py_name)

            try:
                raw_value = value_dict.pop(doc_name)
            except KeyError:
                raise ConfigurationError(
                    f"The config is missing the field `{doc_name}`"
                )

            try:
                parsed_value = uniserde.from_json(raw_value, py_type)
            except uniserde.SerdeError as err:
                raise ConfigurationError(
                    f"`{raw_value}` is not a valid value for `{doc_name}`"
                ) from err

            parsed_fields[py_name] = parsed_value

        # Are superfluous values allowed?
        if raise_on_superfluous and value_dict:
            raise ConfigurationError(
                f"The config contains superfluous fields: `{'`, `'.join(value_dict)}`"
            )

        # Done
        return parsed_fields, value_dict

    @classmethod
    def _load_from_values(
        cls,
        *,
        values: Iterable[tuple[str, Any]],
        case_transform: Callable[[str], str],
    ) -> Self:
        """
        Loads an instance of the config from the provided values.

        ## Raises

        `ConfigurationError`: if any fields are missing, superfluous or of the wrong
            type.
        """
        # Parse the fields
        parsed_fields, unparsed_fields = cls._parse_fields(
            fields=cls._get_fields(),
            values=values,
            case_transform=case_transform,
            raise_on_superfluous=True,
        )
        assert not unparsed_fields, (
            unparsed_fields,
            "Superfluous values present, despite them being disallowed?",
        )

        # Instantiate the class
        self = object.__new__(cls)

        for field_name, field_value in parsed_fields.items():
            setattr(self, field_name, field_value)

        return self

    @classmethod
    def load_from_json(
        cls,
        source: Path | IO[str] | IO[bytes] | Jsonable,
    ) -> Self:
        """
        Loads an instance of the config from the provided JSON source. The
        source can either be a path to a file, an open file-like object, or the
        result of `json.parse`.

        Note that **strings are not treated as paths**, but rather as already
        parsed string values.

        If `source` is a `Path`, and the file doesn't exist, a template will be
        dumped to the location to help out the user.

        ## Raises

        `ConfigurationError`: if any fields are missing, superfluous or of the wrong
            type.
        """

        # Load the JSON
        #
        # If a path was provided but doesn't exist, dump a template
        if isinstance(source, Path):
            try:
                with open(source, "r") as f:
                    raw_values = json5.load(f)

            except FileNotFoundError:
                source.parent.mkdir(parents=True, exist_ok=True)
                with open(source, "w") as f:
                    f.write(cls._create_json_template(cls._get_fields()))

                raise ConfigurationError(
                    f"Could not find the config file at `{source}`. A template has been created for you. Please fill out the values, then try again"
                )

            except ValueError as err:
                raise ConfigurationError(f"`{source}` is not a valid JSON file: {err}")

        # If the source is already parsed, use it as-is
        elif isinstance(source, (type(None), bool, int, float, str, tuple, list, dict)):
            raw_values = source

        # Read & parse file-like objects
        else:
            try:
                raw_values = json5.load(source)

            except ValueError as err:
                raise ConfigurationError(f"`{source}` is not valid JSON: {err}")

        # Make sure the JSON has parsed into a dictionary
        if not isinstance(raw_values, dict):
            raise ConfigurationError("The config JSON must be a dictionary.")

        # Parse the values into an instance
        return cls._load_from_values(
            values=raw_values.items(),
            case_transform=uniserde.case_convert.all_lower_to_camel_case,
        )

    @classmethod
    def load_from_environment(
        cls,
    ) -> Self:
        """
        Loads an instance of the config from the provided JSON source. The
        source can either be a path to a file, an open file-like object, or the
        result of `json.parse`.

        Note that **strings are not treated as paths**, but rather as already
        parsed string values.

        ## Raises

        `ConfigurationError`: if any fields are missing, superfluous or of the wrong
            type.
        """

        raise NotImplementedError("TODO: This function is entirely untested")

        # Fetch all fields from the environment
        raw_values: dict[str, str] = {}

        for py_field_name, field_type in cls._get_fields():
            env_name = py_field_name.upper()

            try:
                raw_value = os.environ[env_name]
            except KeyError:
                raise ConfigurationError(
                    f"There is no environment variable set for `{env_name}`"
                )

            raw_values[env_name] = raw_value

        # Right now all values are strings. Parse them into something more
        # adequate using JSON.
        semi_parsed: dict[str, Any] = {}

        for py_field_name, raw_value in raw_values.items():
            # Keep strings as-is
            if field_type is str:
                semi_parsed[py_field_name] = raw_value
                continue

            # Otherwise drag them through the JSON parser
            try:
                semi_parsed[py_field_name] = json5.loads(raw_value)
            except ValueError as err:
                raise ConfigurationError(f"`{raw_value}` is not a valid value") from err

        # Parse the fields into an instance
        return cls._load_from_values(
            values=semi_parsed.items(),
            case_transform=str.upper,
        )
